//
//  TemplateMaskSurface.swift
//  SLFlops
//
//  Created by Silverlabs on 27/07/21.
//  Copyright © 2021 Silverlabs. All rights reserved.
//

import MetalKit

class TemplateMaskSurface: TemplateBase {
    
    private let device: MTLDevice
    private let channel: SlideShowTemplateConfig.KeyPointFiles.Channel
    
    var texture: MTLTexture?
    var maskTexture: MTLTexture?
    var pipelineState: MTLRenderPipelineState!
    var fragmentFunctionName: String = "textured_mask_fragment"
    let surfaceType: TemplateRenderer.SurfaceType
    
    private var vertexBuffer: MTLBuffer?
    private var indicesBuffer: MTLBuffer?
    private var vertices: [Vertex] = []
    private var indices: [UInt16] = [
        0, 1, 2,
        0, 2, 3 ]
    
    init(device: MTLDevice,
         image: UIImage,
         maskedImage: CIImage?,
         channel: SlideShowTemplateConfig.KeyPointFiles.Channel,
         surfaceType: TemplateRenderer.SurfaceType,
         coordinates: SlideShowTemplateConfig.CoordinatesRow,
         renderSize: CGSize) {
        self.device = device
        self.channel = channel
        self.surfaceType = surfaceType
        
        super.init()
        
        self.texture = try? setTexture(device: device, with: image)
        
        if let maskedImage = maskedImage {
            self.maskTexture = try? setTexture(device: device, with: maskedImage)
        }
        
        prepareVerticesData(from: coordinates, renderSize: renderSize)
        pipelineState = buildPipelineState(device: device)
        buildBuffers()
        
        debugPrint(convertToMetalCoordinates(point: coordinates.topLeft, viewSize: UIScreen.main.bounds.size))
    }
       
    override func render(commandEncoder: MTLRenderCommandEncoder, buffer: MTLCommandBuffer) {
        super.render(commandEncoder: commandEncoder, buffer: buffer)
        
        guard let indexBuffer = indicesBuffer else { return }
        
        commandEncoder.setRenderPipelineState(pipelineState)
        commandEncoder.setVertexBuffer(vertexBuffer, offset: 0, index: 0)
        
        commandEncoder.setFragmentTexture(texture, index: 0)
        commandEncoder.setFragmentTexture(maskTexture, index: 1)
        
        var channelValue = channel.intValue
        commandEncoder.setFragmentBytes(&channelValue, length: MemoryLayout<Int>.stride, index: 0)
        
        commandEncoder.drawIndexedPrimitives(type: .triangle,
                                             indexCount: indices.count,
                                             indexType: .uint16,
                                             indexBuffer: indexBuffer,
                                             indexBufferOffset: 0)
        
    }
    
    deinit {
        texture = nil
        maskTexture = nil
    }
}

// MARK: - TemplateRenderable
extension TemplateMaskSurface: TemplateRenderable {
    
    var vertexFunctionName: String {
        return "vertex_shader"
    }
    
    var vertexDescriptor: MTLVertexDescriptor {
        let vertexDescriptor = MTLVertexDescriptor()
        vertexDescriptor.attributes[0].format = .float3
        vertexDescriptor.attributes[0].offset = 0
        vertexDescriptor.attributes[0].bufferIndex = 0
        
        vertexDescriptor.attributes[1].format = .float4
        vertexDescriptor.attributes[1].offset = MemoryLayout<SIMD3<Float>>.stride
        vertexDescriptor.attributes[1].bufferIndex = 0
        
        vertexDescriptor.attributes[2].format = .float2
        vertexDescriptor.attributes[2].offset = MemoryLayout<SIMD3<Float>>.stride + MemoryLayout<SIMD4<Float>>.stride
        vertexDescriptor.attributes[2].bufferIndex = 0
        
        vertexDescriptor.layouts[0].stride = MemoryLayout<Vertex>.stride
        return vertexDescriptor
    }
}

// MARK: - Texturable
extension TemplateMaskSurface: Texturable {}

private extension TemplateMaskSurface {
    
    func buildBuffers() {
        vertexBuffer =  device.makeBuffer(bytes: vertices, length: vertices.count * MemoryLayout<Vertex>.stride, options: [])
        indicesBuffer = device.makeBuffer(bytes: indices, length: indices.count * MemoryLayout<UInt16>.size, options: [])
    }
    
    func convertToMetalCoordinates(point: CGPoint, viewSize: CGSize) -> simd_float2 {
        let inverseViewSize = CGSize(width: 1.0 / viewSize.width, height: 1.0 / viewSize.height)
        let clipX = Float((2.0 * point.x * inverseViewSize.width) - 1.0)
        let clipY = Float((2.0 * -point.y * inverseViewSize.height) + 1.0)
        return simd_float2(clipX, clipY)
    }
    
    func prepareVerticesData(from coordinates: SlideShowTemplateConfig.CoordinatesRow,
                             renderSize: CGSize) {
        let width: CGFloat = renderSize.width
        let height: CGFloat = renderSize.height

        let topLeft = convertToMetalCoordinates(point: CGPoint(x: coordinates.topLeft.x * width, y: coordinates.topLeft.y * height), viewSize: renderSize)
        let topRight = convertToMetalCoordinates(point: CGPoint(x: coordinates.topRight.x * width, y: coordinates.topRight.y * height), viewSize: renderSize)
        let bottomLeft = convertToMetalCoordinates(point: CGPoint(x: coordinates.bottomLeft.x * width, y: coordinates.bottomLeft.y * height), viewSize: renderSize)
        let bottomRight = convertToMetalCoordinates(point: CGPoint(x: coordinates.bottomRight.x * width, y: coordinates.bottomRight.y * height), viewSize: renderSize)

        vertices = [
            Vertex(position: SIMD3(Float(topLeft.x), Float(topLeft.y), 0), color: SIMD4(1, 0, 0, 1), texture: SIMD2(0, 1)),
            Vertex(position: SIMD3(Float(bottomLeft.x), Float(bottomLeft.y), 0), color: SIMD4(0, 1 , 0, 1), texture: SIMD2(0, 0)),
            Vertex(position: SIMD3(Float(bottomRight.x), Float(bottomRight.y), 0), color: SIMD4(0, 0, 1, 1), texture: SIMD2(1, 0)),
            Vertex(position: SIMD3(Float(topRight.x), Float(topRight.y), 0), color: SIMD4(1, 0, 1, 1), texture: SIMD2(1, 1))
        ]
    }
}

